import re

with open('README.rst') as readme:
    content = readme.read()
    checksum = re.search(r'\|checksum\| replace:: ``(.*)``', content).group(1)
    version = re.search(r'\|version\| replace:: (.*)', content).group(1)
    version_suffix = re.search(r'\|fullversion\| replace:: \|version\|(.*)',
                               content).group(1)
    full_version = version + version_suffix



"""
Tests the setup.py script by running ``pip install gae_installer=={version}``
in a temporarily activated virtual environment.
"""

import os
import unittest
import sys

import version
import test


class TestPostRelease(test.Test):
    def _install(self):
        os.system('pip install gae_installer=={}'.format(version.full_version))


if __name__ == '__main__':
    if sys.version_info.major == 2 and sys.version_info.minor >= 7:
        unittest.main()
    else:
        sys.exit('GAE Installer requires Python 2.7 or higher!')


import os
from distutils.command.build import build
from distutils.command.build_scripts import build_scripts
from distutils.core import setup
import hashlib
import tempfile
import urllib
import zipfile

import version


VERSION = version.full_version if version.full_version.count('.') == 2 else \
    version.full_version.rsplit('.', 1)[0]

GAE_URL = ('https://storage.googleapis.com/appengine-sdks/{0}/'
           'google_appengine_{1}.zip')
GAE_URL_FEATURED = GAE_URL.format('featured', VERSION)
GAE_URL_DEPRECATED = GAE_URL.format('deprecated/{0}'
                                    .format(VERSION.replace('.', '')), VERSION)

GAE_DIR_SCRIPT_NAME = '_get_gae_dir'
SCRIPT_TEMPLATE = 'python `{0}`/`basename $0`.py "$@"'\
    .format(GAE_DIR_SCRIPT_NAME)

script_paths = []

class BuildScripts(build_scripts):
    """Custom build_scripts command"""
    def finalize_options(self):
        global script_paths
        build_scripts.finalize_options(self)
        self.scripts = script_paths


class Build(build):
    """Custom build command"""

    zip_path = os.path.join(tempfile.gettempdir(),
                            'google_appengine_{0}.zip'.format(VERSION))

    def run(self):
        # assert False, (self.build_base, self.build_temp)
        self._get_from_cache_or_download()
        self._populate_files()
        self._populate_scripts()

        build.run(self)

    def _populate_scripts(self):
        """
        Generates script wrapper files for all GAE SDK scripts
        and populates the global script_paths variable with their paths.
        """
        global script_paths
        files = os.listdir(os.path.join(self.build_lib, 'google_appengine'))

        # Exclude the run_tests.py file
        files.remove('run_tests.py')

        scripts_temp = os.path.join(self.build_temp, 'scripts')

        os.makedirs(scripts_temp)

        # Create script for getting the path of the installed GAE SDK
        gae_dir_script_path = os.path.join(scripts_temp, GAE_DIR_SCRIPT_NAME)
        with open(gae_dir_script_path, 'w') as f:
            f.write('python -c "import google.appengine; print google.__file__'
                    '.split(\'/google/\')[0]"')

        script_paths = [gae_dir_script_path]
        for name in files:
            if name.endswith('.py') and name[0] != '_':
                name = name[:-3]
                script_path = os.path.join(scripts_temp, name)
                script_paths.append(script_path)
                with open(script_path, 'w') as f:
                    print 'Generating script file: {0}'.format(script_path)
                    f.write(SCRIPT_TEMPLATE)

    def _populate_files(self):
        """Unzips the downloaded GAE SDK and creates a PTH file"""
        os.makedirs(self.build_lib)
        self._unzip()
        pth_path = os.path.join(self.build_lib, 'google_appengine.pth')
        with open(pth_path, 'w') as f:
            f.write('google_appengine')

    def _get_from_cache_or_download(self):
        """Gets the GAE SDK from cache or downloads it."""
        if os.path.isfile(self.zip_path):
            print('GAE SDK zip found at {0}'.format(self.zip_path))
            if not self._checksum(self.zip_path):
                print('GAE zip checksum doesnt match with {0}!'
                      .format(version.checksum))
                self._download_gae()
        else:
            self._download_gae()

    def _unzip(self):
        """Unzips the GAE SDK"""
        zf = zipfile.ZipFile(self.zip_path)
        print('Extracting {0} to {1}'.format(self.zip_path, self.build_lib))
        zf.extractall(self.build_lib)

    def _download_gae(self):
        """Downloads the GAE SDK"""
        os.makedirs(self.build_base)
        print('Downloading GAE SDK {0} from {1} to {2}'
              .format(VERSION, GAE_URL_FEATURED, self.zip_path))
        print('Please be patient, this can take a while...')
        file_path, response = urllib.urlretrieve(GAE_URL_FEATURED,
                                                 self.zip_path)
        if response.type != 'application/zip':
            print('GAE SDK {0} is deprecated!'.format(VERSION))
            print('Downloading deprecated GAE SDK {0} from {1}'
                  .format(VERSION, GAE_URL_DEPRECATED))
            file_path, response = urllib.urlretrieve(GAE_URL_DEPRECATED,
                                                     self.zip_path)
        if not self._checksum(self.zip_path):
            raise Exception("The downloaded GAE SDK {0} doesn't match the "
                            "SHA1 checksum '{1}'"
                            .format(VERSION, version.checksum))

        print('Download OK')
        return file_path

    def _checksum(self, zip_path):
        """Validates the GAE SDK against a checksum."""
        cs = hashlib.sha1(open(zip_path, 'rb').read()).hexdigest()
        if cs == version.checksum:
            print('Checksum OK')
            return True


setup(
    name='gae_installer',
    version=version.full_version,
    author='Peter Hudec',
    author_email='peterhudec@peterhudec.com',
    description='Google App Engine Installer',
    long_description=open(os.path.join(os.path.dirname(__file__), 'README.rst'))
        .read(),
    keywords='google appengine gae sdk installer',
    url='https://github.com/peterhudec/gae_installer',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Installation/Setup',
    ],
    license='MIT',
    package_data={'': ['*.txt', '*.rst']},
    # If packages is empty, contents of ./build/lib will not be copied!
    packages=['gae_installer'],
    scripts=['if', 'empty', 'BuildScripts', 'will', 'be', 'ignored'],
    cmdclass=dict(build=Build, build_scripts=BuildScripts)
)



"""
Tests the setup.py script by running ``python setup.py install`` in a
temporarily activated virtual environment.
"""

import re
import unittest
import os
import sys
import shutil
import subprocess


VENV_NAME = '_e'

BASE_PATH = os.path.abspath(os.path.dirname(__file__))
VENV_PATH = os.path.join(BASE_PATH, VENV_NAME)
BUILD_PATH = 'build'
ACTIVATE_THIS_PATH = os.path.join(VENV_PATH, 'bin', 'activate_this.py')


def _which(program):
    """Returns path of command executable path."""
    import os
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file


class Test(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.tearDownClass()
        if _which('virtualenv'):
            os.system('virtualenv {0} -p python2.7'.format(VENV_PATH))
            if not os.path.isdir(VENV_PATH):
                sys.exit('Failed to create virtual environment "{0}"!'
                         .format(VENV_PATH))
            cls._activate_venv()
        else:
            sys.exit('Cannot run tests because the "virtualenv" '
                     'command is not installed on your system!'
                     '\nRead installation instructions '
                     'here:\nhttps://virtualenv.pypa.io/en/latest/virtualenv'
                     '.html#installation')

    @classmethod
    def tearDownClass(cls):
        cls._remove_venv()
        cls._remove_build()

    @classmethod
    def _remove_venv(cls):
        """Removes virtual environment"""
        if os.path.isdir(VENV_PATH):
            shutil.rmtree(VENV_PATH)

    @classmethod
    def _remove_build(cls):
        """Removes virtual environment"""
        if os.path.isdir(BUILD_PATH):
            shutil.rmtree(BUILD_PATH)

    @classmethod
    def _activate_venv(cls):
        """
        Activates virtual environment

        http://virtualenv.readthedocs.org/en/latest/virtualenv.html#using-virtualenv-without-bin-python
        """
        execfile(ACTIVATE_THIS_PATH, dict(__file__=ACTIVATE_THIS_PATH))

    def _install(self):
        os.system('python {0} install'
                  .format(os.path.join(BASE_PATH, 'setup.py')))

    def _import_gae(self):
        import google.appengine

    def test_import(self):
        # On Ubuntu, there is a google module in system site-packages
        # which shadows the google module from google_appengine.
        # Therefore we will keep only the virtualenv paths in the pythonpath.
        sys.path = sys.path[:3]

        # GAE import should fail first
        self.assertRaises(ImportError, self._import_gae)

        # After running setup.py,
        self._install()

        # and activating the virtual environment
        self._activate_venv()

        # GAE should not fail
        import google.appengine

        # Patter for elimination of _e/lib/python2.7 and _e/local/lib/python2.7
        # differences in scripts output
        venv_lib_pattern = re.compile(r'(_e/).*(/python)')
        venv_lib_replacement = r'\1...\2'

        # The _get_gae_dir file should exist
        get_gae_dir_path = os.path.join(VENV_PATH, 'bin', '_get_gae_dir')
        self.assertTrue(os.path.isfile(get_gae_dir_path))

        # The _get_gae_command should return the path of the
        # installed google_appengine SDK
        gae_dir = google.appengine.__file__.split('/google/')[0]
        output, error = subprocess.Popen(['_get_gae_dir'],
                                         stderr=subprocess.PIPE,
                                         stdout=subprocess.PIPE,
                                         shell=True).communicate()

        output = venv_lib_pattern.sub(venv_lib_replacement, output.strip())
        gae_dir_clean = venv_lib_pattern.sub(venv_lib_replacement, gae_dir)
        self.assertEquals(output, gae_dir_clean)

        # Skip the run_tests.py file
        original_commands = os.listdir(gae_dir)
        original_commands.remove('run_tests.py')

        # Patter for replacing time in output
        time_pattern = re.compile(r'\d\d:\d\d:\d\d,\d\d\d')

        for command in original_commands:
            if command.endswith('.py') and command[0] != '_':
                original_file = os.path.join(gae_dir, command)
                name = command[:-3]

                self.assertTrue(os.path.isfile(original_file),
                                "File {} doesn't exist!".format(original_file))

                original_output, original_error = subprocess.Popen(
                    ['python', original_file],
                    stderr=subprocess.PIPE,
                    stdout=subprocess.PIPE
                ).communicate()

                self.assertTrue(os.path.isfile(
                    os.path.join(VENV_PATH, 'bin', name)),
                    "File {} doesn't exist!".format(name)
                )

                output, error = subprocess.Popen(
                    [name],
                    stderr=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    shell=True
                ).communicate()

                # Output can contain varying time so we need to eliminate it
                original_output = time_pattern.sub('', original_output)
                original_error = time_pattern.sub('', original_error)
                output = time_pattern.sub('', output)
                error = time_pattern.sub('', error)

                # Eliminate of _e/lib/python2.7 and _e/local/lib/python2.7
                # differences
                original_output = venv_lib_pattern.sub(venv_lib_replacement,
                                                       original_output)
                original_error = venv_lib_pattern.sub(venv_lib_replacement,
                                                      original_error)
                output = venv_lib_pattern.sub(venv_lib_replacement, output)
                error = venv_lib_pattern.sub(venv_lib_replacement, error)

                assert output == original_output
                assert error == original_error

                self.assertEquals(output, original_output,
                                  "Stdouts of {} and {} don't match!"
                                  .format(name, original_file))
                self.assertEquals(error, original_error,
                                  "Stderrs of {} and {} don't match!"
                                  .format(name, original_file))

                ok = output == original_output and error == original_error
                print 'TESTING SCRIPT: {} {}'\
                    .format(name, 'OK' if ok else 'ERROR')


if __name__ == '__main__':
    if sys.version_info.major == 2 and sys.version_info.minor >= 7:
        unittest.main()
    else:
        sys.exit('GAE Installer requires Python 2.7 or higher!')
